import * as sanitizeHtml from 'sanitize-html';
import {h} from 'preact';
import {OnClickEvent} from '@playkit-js/common/dist/hoc/a11y-wrapper';
import {ui} from '@playkit-js/kaltura-player-js';
import {UpperBarManager, SidePanelsManager} from '@playkit-js/ui-managers';
import {decodeString} from './utils';
import {icons} from './components/icons';
import {PluginButton} from './components/plugin-button/plugin-button';
import {LiveTranscript} from './components/live-transcript';
import {getConfigValue, isBoolean, prepareCuePoint} from './utils';
import {LiveTranscriptConfig, PluginStates, HighlightedMap, CuePointData, ItemTypes, CuePoint} from './types';

const {SidePanelModes, SidePanelPositions, ReservedPresetNames} = ui;
const {withText, Text} = KalturaPlayer.ui.preacti18n;

interface TimedMetadataEvent {
  payload: {
    cues: Array<CuePoint>;
  };
}

export class LiveTranscriptPlugin extends KalturaPlayer.core.BasePlugin {
  static defaultConfig: LiveTranscriptConfig = {
    expandMode: SidePanelModes.ALONGSIDE,
    expandOnFirstPlay: true,
    showTime: true,
    position: SidePanelPositions.RIGHT,
    scrollOffset: 0,
    searchDebounceTimeout: 250,
    searchNextPrevDebounceTimeout: 100,
    downloadDisabled: false,
    printDisabled: false
  };
  private _activeCaptionMapId: string = '';
  private _activeCuePointsMap: HighlightedMap = {};
  private _activeTextTrack: any = null;
  private _captionMap: Map<string, Array<CuePointData>> = new Map();
  private _isLoading = false;
  private _hasError = false;
  private _triggeredByKeyboard = false;

  private _liveTranscriptPanel = -1;
  private _liveTranscriptIcon = -1;
  private _pluginState: PluginStates | null = null;
  private _pluginButtonRef: HTMLButtonElement | null = null;
  private _data: CuePointData[] = [];
  private _isSafari: boolean = false;

  constructor(name: string, player: KalturaPlayerTypes.Player, config: LiveTranscriptConfig) {
    super(name, player, config);

    var ua = navigator.userAgent.toLowerCase();

    // Check if it's Safari
    if (ua.includes('safari') && !ua.includes('chrome') && !ua.includes('android')) {
      this._isSafari = true;
    }
  }

  get sidePanelsManager() {
    return this.player.getService('sidePanelsManager') as SidePanelsManager | undefined;
  }

  get upperBarManager() {
    return this.player.getService('upperBarManager') as UpperBarManager | undefined;
  }

  get cuePointManager() {
    return this.player.getService('kalturaCuepoints') as any;
  }

  loadMedia(): void {
    this.player.ready().then(()=> {
      if (!this.sidePanelsManager || !this.upperBarManager) {
        this.logger.warn("sidePanelsManager or upperBarManager haven't registered");
        return;
      }

      this._initListeners();

      this.eventManager.listenOnce(
          this.player,
          this.player.Event.Core.FIRST_PLAYING,
          () => {
            if (!this.player.isLive()) {
              this.logger.warn("Live transcript plugin is not supported for non-live entries");
              return;
            }

            this._addTranscriptItem();
          },
      );
    });
  }

  private async setActiveTextTrack() {
    const allTextTracks = this._getTextTracks();
    const activePlayerTextTrack = allTextTracks.find((track: any) => track.active);
    const TextTracks = Object.values(this.player.getVideoElement().textTracks);
    const userLanguage = navigator.language;
    const shortLanguage = userLanguage.substring(0, 2);

    // first we check if we found track with "active" = true, and it is actually not "off"
    // in that case we search for the track with the same language
    if (typeof activePlayerTextTrack !== 'undefined' && activePlayerTextTrack.language !== "off") {
      this.logger.info("trying to set activeTextTrack, condition 'undefined' or 'off'");

      this._activeTextTrack = TextTracks.find((track: any) => {
        return track.language === activePlayerTextTrack.language;
      });
    }

    // if nothing set above - try to find track with user's language
    if (!this._activeTextTrack || typeof this._activeTextTrack === 'undefined' || this._activeTextTrack?.language === '') {
      this.logger.info("trying to set activeTextTrack 2, condition 'undefined' or ''");
      this.logger.info(this._activeTextTrack);

      this._activeTextTrack = TextTracks.find((track: any) => {
        return track.language === shortLanguage;
      });
    }

    // in some cases, if navigator.language is not found among text tracks
    // we should fall back to english in there is such text track
    if (this._activeTextTrack?.language === '') {
      this.logger.info("falling back to EN text track");

      this._activeTextTrack = TextTracks.find((track: any) => {
        return track.language === 'en';
      });
    }

    // if nothing set above - fallback to the first track in the list
    if (!this._activeTextTrack || typeof this._activeTextTrack === 'undefined') {
      if (!this._isSafari) {
        this._activeTextTrack = TextTracks[0];
      } else {
        this._activeTextTrack = TextTracks.find((track: any) => {
          return track.kind === "metadata" || track.active === true;
        });
      }
    }

    if (this._activeTextTrack.mode === "disabled") { // fix for Safari: do not hide CC if it is in the "showing" mode already
      this._activeTextTrack.mode = "hidden"; // this override allows for live transcript to work even if Closed Captions are turned off
    }

    this.logger.info("activeTextTrack set:");
    this.logger.info(this._activeTextTrack);
  }

  private _initListeners(): void {
    const eventToListen = this._isSafari ? this.player.Event.TEXT_CUE_CHANGED : this.player.Event.PLAYING;

    this.eventManager.listenOnce(this.player, eventToListen, () => {
      this.logger.info(eventToListen + " event happened");
      this.logger.info("Text tracks:");
      this.logger.info(this._getTextTracks());

      // Initialize a set to track unique cue content (tracking content + timing)
      const uniqueCueSet = new Set<string>();

      if (this._getTextTracks().length) {
        this.eventManager.listen(this.player, this.player.Event.TIMED_METADATA_CHANGE, this._onTimedMetadataChange);
        this.eventManager.listen(this.player, this.player.Event.TIMED_METADATA_ADDED, this._onTimedMetadataAdded);
        this.eventManager.listen(this.player, this.player.Event.TEXT_TRACK_CHANGED, this._handleLanguageChange);

        this.setActiveTextTrack().then(() => {
          if (!this._isSafari) {
            this._activeTextTrack.addEventListener('cuechange', (e: Event) => {
              const target = e.currentTarget as TextTrack;
              const activeCues = target.activeCues as TextTrackCueList;

              this.logger.info('cuechange event happened for active text track');
              this.logger.info('activeCues:');
              this.logger.info(activeCues);

              if (activeCues && target.activeCues !== target.cues) {
                this.processActiveCues(activeCues, uniqueCueSet);
              }
            });
          } else {
            this.eventManager.listen(this.player, this.player.Event.TEXT_CUE_CHANGED, (e: any) => {
              const activeCues = e.payload?.cues as TextTrackCueList;

              this.logger.info('TEXT_CUE_CHANGED event happened for active text track');
              this.logger.info('activeCues:');
              this.logger.info(activeCues);

              this.processActiveCues(activeCues, uniqueCueSet);
            });
          }
        });
      }
    });
  }

  private processActiveCues(activeCues: TextTrackCueList, uniqueCueSet: Set<string>) {
    const additionalData: CuePointData[] = Object.values(activeCues)
      .filter((textTrack) => {
        const cue = textTrack as VTTCue;

        if (!cue.text || typeof cue.text !== 'string') {
          this.logger.warn(`Skipping cue with invalid text: ${cue}`);
          return false; // Skip cues with undefined or non-string text
        }

        const cueText = cue.text.trim();
        const cueStartTime = cue.startTime;
        const cueEndTime = cue.endTime;

        // Split the cue text by newlines to handle multi-line cues
        let pieces = cueText ? cueText.split('\n') : [cueText];

        // Handle single-line cues or when there’s no previous cue
        if (pieces.length === 1) {
          // Check the last five elements in _data
          const lastElements = this._data.slice(-5);
          let shouldAddCue = true; // Flag to determine if we should add the new cue

          for (let i = 0; i < lastElements.length; i++) {
            const lastCue = lastElements[i];

            // Check if the last cue text is fully equal to the current cue text
            if (lastCue.text === cueText) {
              this.logger.info(`Duplicate cue skipped: ${cueText}`);
              return false; // Filter out this cue as a duplicate
            }

            // Check if the current cue text is partially equal to the last cue text
            if (cueText.includes(lastCue.text)) {
              this.logger.info(`Updating last cue to: ${cueText}`);
              // Replace the last cue's text with the current cue text

              // Update the original last cue in this._data
              const index = this._data.indexOf(lastCue);
              if (index !== -1) {
                this._data[index].text = cueText; // Update the text directly in this._data

                this._updateLiveTranscriptPanel();
              }

              // Since we are updating an existing cue, we should not add a new one
              shouldAddCue = false;
              break; // Exit the loop once we update
            }
          }

          // If we haven't replaced any last cue, check for uniqueness
          if (shouldAddCue) {
            const cueKey = `${cueText}-${cueStartTime}-${cueEndTime}`;

            if (!uniqueCueSet.has(cueKey)) {
              uniqueCueSet.add(cueKey);
              return true; // Indicate that this cue should be added
            }

            this.logger.info(`Duplicate cue skipped: ${cueKey}`);
          }

          return false; // Skip adding a new entry
        }

        let currentFirstLine = pieces[0];
        let lastLine = this._data.at(-1)?.text ? this._data.at(-1)?.text.split('\n') : [this._data.at(-1)?.text];
        const result = lastLine?.at(-1) !== currentFirstLine && lastLine?.at(0) !== currentFirstLine;
        this.logger.info(lastLine);
        this.logger.info(pieces);

        // If the first line matches the last line from the previous cue, skip it
        if (!result) {
          this.logger.info(`Multi-line cue overlap detected. Skipping line: ${currentFirstLine}`);
          return false;
        }

        // If the cue is valid, create a unique key and add it to the set
        const cueKey = `${cueText}-${cueStartTime}-${cueEndTime}`;
        if (!uniqueCueSet.has(cueKey)) {
          uniqueCueSet.add(cueKey);
          return true;
        }

        return false; // Skip if it's a duplicate
      })
      .map((cue: any) => ({
        id: (KalturaPlayer.core.utils as any).Generator.guid(), // Generate unique IDs
        startTime: cue.startTime,
        endTime: cue.endTime,
        text: cue.text
      }));

    // If there are new cues, update the _data array
    if (additionalData.length > 0) {
      this._data = [...this._data, ...additionalData.reverse()];
    }

    this._updateLiveTranscriptPanel();
  }

  private _initLoading = () => {
    this._isLoading = false;
    this._hasError = false;
    this._updateLiveTranscriptPanel();
  };

  private _handleLanguageChange = () => {
    this.logger.info("TEXT_TRACK_CHANGED event happened");
    this.logger.info("Text tracks:");
    this.logger.info(this._getTextTracks());

    this._activeCaptionMapId = this._getCaptionMapId();
    this._initLoading();
    this.setActiveTextTrack();
  };

  private _updateLiveTranscriptPanel() {
    if (this._liveTranscriptPanel) {
      this.sidePanelsManager?.update(this._liveTranscriptPanel);
    }
  }

  private _onTimedMetadataAdded = ({payload}: TimedMetadataEvent) => {
    const captionData: CuePointData[] = [];
    payload.cues.forEach((cue: CuePoint) => {
      if (cue.metadata.cuePointType === ItemTypes.Caption) {
        captionData.push(prepareCuePoint(cue));
      }
    });
    if (captionData.length && this.player.isLive()) {
      this._addCaptionData(captionData);
      this._addTranscriptItem();
    }
  };

  private _onTimedMetadataChange = ({payload}: TimedMetadataEvent) => {
    const transcriptCuePoints: Array<CuePoint> = payload.cues
      .filter((cue: CuePoint) => {
        return cue.metadata.cuePointType === ItemTypes.Caption;
      })
      .filter((cue, index, array) => {
        // filter out captions that has endTime eq to next caption startTime
        const nextCue = array[index + 1];
        return !nextCue || cue.endTime !== nextCue.startTime;
      });
    this._activeCuePointsMap = {};
    transcriptCuePoints.forEach(cue => {
      this._activeCuePointsMap[cue.id] = true;
    });
    this._updateLiveTranscriptPanel();
  };

  private _addCaptionData = (newData: CuePointData[]) => {
    this._activeCaptionMapId = this._getCaptionMapId();
    this._captionMap.set(this._activeCaptionMapId, this._sanitizeCaptions(newData));
    this._isLoading = false;
    this._updateLiveTranscriptPanel();
  };

  private _getTextTracks = () => {
    return this.player.getTracks(this.player.Track.TEXT) || [];
  };

  private _getCaptionMapId = (): string => {
    const allTextTracks = this._getTextTracks();
    const activeTextTrack = allTextTracks.find(track => track.active);
    if (activeTextTrack?.language === 'off') {
      // use 1st captions from text-track list
      return `${allTextTracks[0]?.language}-${allTextTracks[0]?.label}`;
    }
    return `${activeTextTrack?.language}-${activeTextTrack?.label}`;
  };

  private _activatePlugin = () => {
    this.ready.then(() => {
      this.sidePanelsManager?.activateItem(this._liveTranscriptPanel);
      this._pluginState = PluginStates.OPENED;
      this.upperBarManager?.update(this._liveTranscriptIcon);
    });
  };

  private _deactivatePlugin = () => {
    this.ready.then(() => {
      this.sidePanelsManager?.deactivateItem(this._liveTranscriptPanel);
      this.upperBarManager?.update(this._liveTranscriptIcon);
    });
  };

  private _isPluginActive = () => {
    return this.sidePanelsManager!.isItemActive(this._liveTranscriptPanel);
  };

  private _handleClickOnPluginIcon = (e: OnClickEvent, byKeyboard?: boolean) => {
    if (this._isPluginActive()) {
      this._triggeredByKeyboard = false;
      this._deactivatePlugin();
      this._pluginState = PluginStates.CLOSED;
    } else {
      this._triggeredByKeyboard = Boolean(byKeyboard);
      this._activatePlugin();
    }
  };

  private _sanitizeCaptions = (data: CuePointData[]) => {
    return data.map(caption => ({
      ...caption,
      text: decodeString(
        sanitizeHtml(caption.text || '', {
          allowedAttributes: {},
          allowedTags: []
        })
      )
    }));
  };

  private _addTranscriptItem(): void {
    if (Math.max(this._liveTranscriptPanel, this._liveTranscriptIcon) > 0) {
      // transcript panel or icon already exist
      return;
    }

    const {
      expandMode,
      position,
      expandOnFirstPlay,
      showTime,
      scrollOffset,
      searchDebounceTimeout,
      searchNextPrevDebounceTimeout,
    } = this.config;
    this._liveTranscriptPanel = this.sidePanelsManager!.add({
      label: 'Live Transcript',
      panelComponent: () => {
        return (
          <LiveTranscript
            showTime={getConfigValue(showTime, isBoolean, true)}
            scrollOffset={getConfigValue(scrollOffset, Number.isInteger, 0)}
            searchDebounceTimeout={getConfigValue(searchDebounceTimeout, Number.isInteger, 250)}
            searchNextPrevDebounceTimeout={getConfigValue(searchNextPrevDebounceTimeout, Number.isInteger, 100)}
            highlightedMap={this._activeCuePointsMap}
            onSeek={this._seekTo}
            onItemClicked={this._seekTo}
            expandMode={this.config.expandMode}
            captions={this._data}
            isLoading={false}
            hasError={this._hasError}
            onRetryLoad={this._updateLiveTranscriptPanel}
            currentTime={this.player.currentTime}
            videoDuration={this.player.duration}
            kitchenSinkActive={this._isPluginActive()}
            toggledWithEnter={this._triggeredByKeyboard}
            onClose={this._handleClose}
            downloadDisabled={true}
            onDownload={this._handleDownload}
            printDisabled={true}
            onPrint={this._handlePrint}
          />
        );
      },
      presets: [ReservedPresetNames.Playback, ReservedPresetNames.Live, ReservedPresetNames.Ads],
      position: position,
      expandMode: expandMode === SidePanelModes.ALONGSIDE ? SidePanelModes.ALONGSIDE : SidePanelModes.OVER,
      onDeactivate: this._deactivatePlugin
    }) as number;
    const translates = {
      showTranscript: <Text id="transcript.show_plugin">Show Live Transcript</Text>,
      hideTranscript: <Text id="transcript.hide_plugin">Hide Live Transcript</Text>
    };
    this._liveTranscriptIcon = this.upperBarManager!.add({
      label: 'Live Transcript',
      svgIcon: {path: icons.PLUGIN_ICON, viewBox: `0 0 ${icons.BigSize} ${icons.BigSize}`},
      onClick: this._handleClickOnPluginIcon as () => void,
      component: withText(translates)((props: {showTranscript: string; hideTranscript: string}) => {
        const isActive = this._isPluginActive();
        const label = isActive ? props.hideTranscript : props.showTranscript;
        return (
          <PluginButton
            isActive={isActive}
            id="liveTranscript-icon"
            label={label}
            icon={icons.PLUGIN_ICON}
            dataTestId="liveTranscript_pluginButton"
            setRef={this._setPluginButtonRef}
          />
        );
      })
    }) as number;

    if ((expandOnFirstPlay && !this._pluginState) || this._pluginState === PluginStates.OPENED) {
      this._activatePlugin();
    }
  }

  private _setPluginButtonRef = (ref: HTMLButtonElement | null) => {
    this._pluginButtonRef = ref;
  };

  private _seekTo = (time: number) => {
    this.player.currentTime = time;
  };

  private _handleDownload = () => {
    //
  };

  private _handlePrint = () => {
    //
  };

  private _handleClose = (e: OnClickEvent, byKeyboard: boolean) => {
    if (byKeyboard) {
      this._pluginButtonRef?.focus();
    }
    this._deactivatePlugin();
    this._pluginState = PluginStates.CLOSED;
  };

  static isValid(): boolean {
    return true;
  }

  reset(): void {
    this.eventManager.removeAll();
    if (Math.max(this._liveTranscriptPanel, this._liveTranscriptIcon) > 0) {
      this.sidePanelsManager?.remove(this._liveTranscriptPanel);
      this.upperBarManager!.remove(this._liveTranscriptIcon);
      this._liveTranscriptPanel = -1;
      this._liveTranscriptIcon = -1;
      this._pluginButtonRef = null;
    }
    this._captionMap = new Map();
    this._activeCaptionMapId = '';
    this._isLoading = false;
    this._hasError = false;
    this._triggeredByKeyboard = false;
  }

  destroy(): void {
    this.reset();
  }
}
